#include "torero.h"

int torero_init( int argc,
		 char ***argv )
{
  /* set defaults */
  internal_data_set_start_code ("AAA");
  internal_data_set_from_dir ("./");
  internal_data_set_to_dir ("./");
  
  internal_data_set_verbose_level(0);

  internal_data_set_up_files_tree ();
  
  if ( argc > 1 ) {
    (*argv)++; /* we don't interested in program name */
    argc--;

    do {
      if ( strlen (**argv) < 2 ) { /* it is not command */
	torero_printf ("Error during processing commands", 0);
	exit (1);
      }
      if ( (***argv == '-') && (strlen (**argv) == 2) ) { /* we have short command with space*/
	switch ( *((**argv)+1) ) {
	case 'f' : /* set from dir */
	  (*argv)++;
	  argc--;
	  if ( argc  == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_from_dir (**argv);
	  break;

	case 't' : /* set to dir */
	  (*argv)++;
	  argc--;
	  if ( argc  == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_to_dir (**argv);
	  break;

	case 'v' : /* set verbose level */
	  (*argv)++;
	  argc--;
	  if ( argc  == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_verbose_level (atoi (**argv));
	  break;
	  
	case 'h' : /* printf help and exit */
	  torero_print_help ();
	  exit (0);
	  break;

	default :
	  torero_printf ("Error during processing commands", 0);
	  exit (1);
	}
      } else if ( (***argv == '-') 
		  && (*((**argv)+1) != '-') 
		  && (strlen (**argv) > 2) )	{ /* we have short command without space */
	switch ( *((**argv)+1) ) {
	case 'f' : /* set from dir */
	  internal_data_set_from_dir (((**argv)+2));
	  break;
	  
	case 't' : /* set to dir */
	  internal_data_set_to_dir (((**argv)+2));
	  break;
	  
	case 'v' : /* set verbose level */
	  internal_data_set_verbose_level (atoi (((**argv)+2)));
	  break;

	case 'h' : /* print help end exit */
	  torero_print_help ();
	  exit (0);

	default :
	  torero_printf ("Error during processing commands", 0);
	  exit (1); 
	}
      } else if ( (***argv == '-') && *((**argv)+1) == '-') { /* we have long command */
	;
	if ( strcmp (**argv, "--from-directory") == 0 ) {
	  (*argv)++;
	  argc--;
	  if ( argc == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_from_dir (**argv);
	} else if ( strcmp (**argv, "--to-directory") == 0 ) {
	  (*argv)++;
	  argc--;
	  if ( argc == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_to_dir (**argv);
	} else if ( strcmp (**argv, "--verbose-level") == 0 ) {
	  (*argv)++;
	  argc--;
	  if ( argc == 0 ) { /* there lacks folowing string in command */
	    torero_printf ("Error during processing commands", 0);
	    exit (1);
	  }
	  internal_data_set_verbose_level (atoi (((**argv)+2)));
	} else if ( strcmp (**argv, "--help") == 0 ) { /* print help and exit */
	  torero_print_help ();
	  exit (0);
	} else if ( strcmp (**argv, "--version") == 0 ) { /* print version number and exit */
	  torero_print_version_num ();
	  exit (0);
	}
      } else if ( strlen (**argv) == 3 
		  && isalpha (*(**argv+0))
		  && isalpha (*(**argv+1))
		  && isalpha (*(**argv+2)) ) { /* we have start code */

	char start_code[4];

	/* we need upper cases in codes */
	start_code[0] = toupper ((*(**argv+0)));
	start_code[1] = toupper ((*(**argv+1)));
	start_code[2] = toupper ((*(**argv+2)));
	start_code[3] = '\0';

	internal_data_set_start_code (start_code);

      } else { /* we have an error in command line */
	torero_printf ("Error during processing commands", 0);
	exit (1);
      }
      
      (*argv)++; /* go to next string */
      argc--;
    } while ( argc > 0 );
  }

  return 0;
}

void torero_printf( char message[], 
		    int message_level )
{
  int verbose_level;

  verbose_level = internal_data_get_verbose_level ();

  if ( message_level <= verbose_level ) {
    printf ("torero: %s\n", message);
  }
}

void torero_print_help( void )
{
  torero_print_version_num ();
  printf ("Usage: torero [OPTION...] [START CODE...]\n");
  printf ("Concatenating corrida export files to database\n");
  printf ("\n");
  printf ("Default start code is AAA.\n");
  printf ("\n");
  printf ("Options:\n");
  printf (" -f, --from-directory  process files from directory (default ./)\n");
  printf (" -t, --to-directory    write results to directory (default ./)\n");
  printf (" -v, --verbose-level   set verbose level (possible values: 0, 1, 2,\n");
  printf ("                       default is 0)\n");
  printf (" -h, --help            print help and exit\n");
  printf ("     --version         print version number and exit\n");
  printf ("\n");
  printf ("Examples:\n");
  printf (" torero\n");
  printf ("	Concatenate files from current working directory,\n");
  printf ("	write results to current directory, start with code AAA\n");
  printf ("\n");
  printf (" torero -f /home/jurmc/meteors2001 CAB\n");
  printf ("	Concatenate files form /home/jurmc/meteors2001,\n");
  printf ("	write results to current directory, start with code CAB\n");
  printf ("\n");
  printf (" torero -t ../\n");
  printf ("	Concatenate files from current working directory,\n");
  printf ("	write results to ../, start with code AAA\n");
}

void torero_print_version_num( void )
{
  printf ("Torero ver.: %s\n", TORERO_VER);
}

gint compare_files( gconstpointer node1,
		    gconstpointer node2 )
{
  gchar year_str_1[3];
  gchar month_str_1[3];
  gchar day_str_1[3];
  gchar observer_1[6];
  gint year_1, month_1, day_1;

  gchar year_str_2[3];
  gchar month_str_2[3];
  gchar day_str_2[3];
  gchar observer_2[6];
  gint year_2, month_2, day_2;

  /* node1 */
  strncpy (year_str_1, node1, 2); year_str_1[2] = '\0';
  strncpy (month_str_1, node1+2, 2); month_str_1[2] = '\0';
  strncpy (day_str_1, node1+4, 2); day_str_1[2] = '\0';
  strncpy (observer_1, node1+7, 5); observer_1[5] = '\0';
  
  year_1 = atoi (year_str_1);
  month_1 = atoi (month_str_1);
  day_1 = atoi (day_str_1);

  /* node2 */
  strncpy (year_str_2, node2, 2); year_str_2[2] = '\0';
  strncpy (month_str_2, node2+2, 2), month_str_2[2] = '\0';
  strncpy (day_str_2, node2+4, 2); day_str_2[2] = '\0';
  strncpy (observer_2, node2+7, 5); observer_2[5] = '\0';

  year_2 = atoi (year_str_2);
  month_2 = atoi (month_str_2);
  day_2 = atoi (day_str_2);

  if ( year_1 < year_2 ) {
    return -1;
  } else if ( year_1 > year_2 ) {
    return 1;
  } else if ( year_1 == year_2 ) {
    if ( month_1 < month_2 ) {
      return -1;
    } else if ( month_1 > month_2 ) {
      return 1;
    } else if ( month_1 == month_2 ) {
      if ( day_1 < day_2 ) {
	return -1;
      } else if ( day_1 > day_2 ) {
	return 1;
      } else if ( day_1 == day_2 ) {
	return strcmp (observer_1, observer_2);
      }
    }
  }

  return 0; /* nodes are equal */ 
}

gboolean concatenate_files( gpointer key,
			    gpointer val, /* it is tree_node*/
			    gpointer data )
{
  tree_node *node_val;
  static FILE *heads_base_fp = NULL;
  static FILE *coords_base_fp = NULL;
  static FILE *notpl_base_fp = NULL;

  node_val = (tree_node *) val;

  /* base files initialization (only once) */
  if ( heads_base_fp == NULL ) {
    heads_base_fp = internal_data_set_up_heads_base_file ();
  }
   
  if ( coords_base_fp == NULL ) {
    coords_base_fp = internal_data_set_up_coords_base_file ();
  }

  if ( notpl_base_fp == NULL ) {
    notpl_base_fp = internal_data_set_up_notpl_base_file ();
  }

  {
    gchar message[MAX_TORERO_STRLINE];
    gchar full_filename[MAX_TORERO_STRLINE];
    FILE *heads_for_reading_fp = NULL;
    
    gchar head_line[MAX_TORERO_STRLINE];
    gchar coords_line[MAX_TORERO_STRLINE];
    gchar notpl_line[MAX_TORERO_STRLINE];

    sprintf (message, "processing node: %s", key);
    torero_printf (message, 1);

    /* open heads file for reading */
    sprintf (full_filename, "%s%s",
	     internal_data_get_from_dir (),
	     node_val->heads_filename);
    sprintf (message, "opening file for reading: %s",
	     full_filename);
    torero_printf (message, 2);
    heads_for_reading_fp = fopen (full_filename, "r");
    if ( heads_for_reading_fp == NULL ) { /* have we really opened the file */
      sprintf (message, "ERROR: unable to open file '%s' for reading", 
	       full_filename);
      torero_printf (message, 0);
      exit (1);
    }

    while ( get_heads_line (heads_for_reading_fp, head_line) != NULL ) {
      gchar *code;
      gchar old_code[4]; /* code from read file */

      code = internal_data_get_next_code ();
      replace_code_in_heads_line (head_line, code, old_code);

      fprintf (heads_base_fp, "%s\n", head_line);

      /* print to coords base file */
      {
	FILE *coords_for_reading_fp = NULL;
	gchar coords_full_filename[MAX_TORERO_STRLINE];

	/* open coords file for reading */
	sprintf (coords_full_filename, "%s%s",
		 internal_data_get_from_dir (),
		 node_val->coords_filename);
	sprintf (message, "opening file for reading: %s",
		 coords_full_filename);
	torero_printf (message, 2);
	coords_for_reading_fp = fopen (coords_full_filename, "r");
	if ( coords_for_reading_fp == NULL ) { /* have we really opened the file */
	  sprintf (message, "ERROR: unable to open file '%s' for reading",
		   coords_full_filename);
	  torero_printf (message, 0);
	  exit (1);
	}

	while ( get_coords_line (coords_for_reading_fp, coords_line) != NULL ) {
	  if ( is_current_code_in_coords_line (coords_line, old_code) != FALSE ) {
	    replace_code_in_coords_line (coords_line, code);
	    fprintf (coords_base_fp, "%s\n", coords_line);
	  }
	}

	/* close coords file for reading */
	if ( coords_for_reading_fp != NULL ) {
	  sprintf (message, "closing file for reading: %s",
		   coords_full_filename);
	  torero_printf (message, 2);
	  fclose (coords_for_reading_fp);
	}
      }

      /* printf to notpl base file */
      {
	FILE *notpl_for_reading_fp = NULL;
	gchar notpl_full_filename[MAX_TORERO_STRLINE];

	/* open notpl file for reading */
	sprintf (notpl_full_filename, "%s%s",
		 internal_data_get_from_dir (),
		 node_val->notpl_filename);
	sprintf (message, "opening file for reading: %s",
		 notpl_full_filename);
	torero_printf (message, 2);
	notpl_for_reading_fp = fopen (notpl_full_filename, "r");
	if ( notpl_for_reading_fp == NULL ) { /* have we really opened the file */
	  sprintf (message, "ERROR: unable to open file '%s' for reading",
		   notpl_full_filename);
	  torero_printf (message, 0);
	  exit (1);
	}
	
	while ( get_notpl_line (notpl_for_reading_fp, notpl_line) != NULL ) {
	  if ( is_current_code_in_notpl_line (notpl_line, old_code) != FALSE ) {
	    replace_code_in_notpl_line (notpl_line, code);
	    fprintf (notpl_base_fp, "%s\n", notpl_line);
	  }
	}

	/* close notpl file for reading */
	if ( notpl_for_reading_fp != NULL ) {
	  sprintf (message, "closing file for reading: %s", 
		   notpl_full_filename);
	  torero_printf (message, 2);
	  fclose (notpl_for_reading_fp);
	}
      }
    }

    /* close heads file for reading */
    if ( heads_for_reading_fp != NULL ) { 
      fclose (heads_for_reading_fp);
    }
  }

  return FALSE; /* allow continue traversing */
}

tree_node *tree_node_new (gchar heads_filename[],
			  gchar coords_filename[],
			  gchar notpl_filename[])
{
  tree_node *new_node;

  new_node = g_new (tree_node, 1);

  new_node->heads_filename = g_strdup (heads_filename);
  new_node->coords_filename = g_strdup (coords_filename);
  new_node->notpl_filename = g_strdup (notpl_filename);

  return new_node;
}

void add_file_to_tree( gchar *heads_filename )
{
  static GTree *tree;
  gchar message[MAX_TORERO_STRLINE];

  tree = internal_data_get_files_tree ();

  /* add file name only if it is heads file */
  if ( strlen (heads_filename) == 22 ) {
    

    if ( strcmp (heads_filename+13, "heads.txt") == 0 ) {
      gchar coords_filename[MAX_TORERO_FILENAME];
      gchar notpl_filename[MAX_TORERO_FILENAME];

      /* name of coords file */
      strcpy (coords_filename, heads_filename);
      sprintf (coords_filename+13, "coords.txt");

      /* name of notpl file */
      strcpy (notpl_filename, heads_filename);
      sprintf (notpl_filename+13, "notpl.txt");

      sprintf (message, "adding node: %s", heads_filename);
      torero_printf (message, 2);
      g_tree_insert (tree, g_strdup (heads_filename), 
		     tree_node_new (heads_filename,
				    coords_filename,
				    notpl_filename));
    }
  }
}

int fgetl( FILE *file, gchar *line, int max_strlen )
{
  gchar c;
  gint line_len = 0;

  sprintf (line, "\0");

  while ( !(feof (file))
          && (c = fgetc (file)) != '\n'
          && line_len < max_strlen ) {
    line[line_len] = c;
    line_len++;
  }

  line[line_len] = '\0';

  return line_len;
}

gchar *get_heads_line( FILE *fp, gchar head_line[] )
{
  gint len;

  len = fgetl (fp, head_line, MAX_TORERO_STRLINE-1);

  if ( len < 75 ) {
    return NULL;
  }

  return head_line;
}

gchar *get_coords_line( FILE *fp, gchar coords_line[] )
{
  gint len;
  
  len = fgetl (fp, coords_line, MAX_TORERO_STRLINE-1);

  if ( len < 70 ) {
    return NULL;
  } 

  return coords_line;
}

gchar *get_notpl_line( FILE *fp, gchar notpl_line[] )
{
  gint len;
  
  len = fgetl (fp, notpl_line, MAX_TORERO_STRLINE-1);

  if ( len < 43 ) {
    return NULL;
  } 

  return notpl_line;
}

gboolean is_current_code_in_coords_line( gchar *coords_line, 
					 gchar *old_code )
{
  if ( coords_line[67] == old_code[0] 
       && coords_line[68] == old_code[1]
       && coords_line[69] == old_code[2] ) {
    return TRUE;
  }

  return FALSE;
}

gboolean is_current_code_in_notpl_line( gchar *notpl_line,
					gchar *old_code )
{
  if ( notpl_line[0] == old_code[0] 
       && notpl_line[1] == old_code[1]
       && notpl_line[2] == old_code[2] ) {
    return TRUE;
  }

  return FALSE;
}

void replace_code_in_heads_line( gchar *head_line, 
				 gchar *code,
				 gchar *old_code_bin)
{
  old_code_bin[0] = head_line[0];
  old_code_bin[1] = head_line[1];
  old_code_bin[2] = head_line[2];
  old_code_bin[3] = '\0';
  
  head_line[0] = code[0];
  head_line[1] = code[1];
  head_line[2] = code[2];
}

void replace_code_in_coords_line( gchar *coords_line,
			     gchar *code )
{
  coords_line[67] = code[0];
  coords_line[68] = code[1];
  coords_line[69] = code[2];
}

void replace_code_in_notpl_line( gchar *notpl_line,
				 gchar *code)
{
  notpl_line[0] = code[0];
  notpl_line[1] = code[1];
  notpl_line[2] = code[2];
}

int main( int   argc,
	  char *argv[] )
{
  torero_init (argc, &argv );
 
  /* directories tests, we should work with it, add a better tests */
  if ( (g_file_test (internal_data_get_from_dir (),
		    G_FILE_TEST_IS_DIR)) != TRUE ) {
    torero_printf ("You specified wrong 'from dir'", 0);
    exit (1);
  }

  if ( (g_file_test (internal_data_get_to_dir (),
		     G_FILE_TEST_IS_DIR)) != TRUE ) {
    torero_printf ("You specified wrong 'to dir'", 0);
    exit (1);
  }

  /* concatenating files */
  {
    GDir *from_dir;
    GError *file_error = NULL;
    gchar *file_name;
    gchar message[MAX_TORERO_STRLINE];

    from_dir = g_dir_open (internal_data_get_from_dir (),
			   0, &file_error);

    while ( (file_name = g_strdup(g_dir_read_name (from_dir))) != NULL ) {
      add_file_to_tree (file_name);
      g_free (file_name);
      
    }
    g_dir_close (from_dir);

    sprintf (message, "counted nodes: %d", 
	     g_tree_nnodes (internal_data_get_files_tree ()));
    torero_printf (message, 0);
    g_tree_foreach (internal_data_get_files_tree (),
		    concatenate_files,
		    NULL);

  }

  internal_data_destroy ();

  return 0;
}
